use std::sync::mpsc::Sender;

use color_eyre::eyre;
use color_eyre::Result;
use reqwest::header;
use reqwest::Client;
use tokio::runtime;
use tokio::sync::mpsc::UnboundedReceiver;

use crate::types::{Event, Message, NtfyMessage};
use crate::APP_USER_AGENT;

#[tracing::instrument]
async fn send_ntfy(client: &Client, data: &NtfyMessage) -> Result<()> {
    let body = serde_json::to_string(data)?;

    let res = client
        .post(&data.url)
        .headers({
            let mut headers = header::HeaderMap::new();

            let auth = format!("Bearer {}", data.token);
            let mut auth = header::HeaderValue::from_str(&auth)?;
            auth.set_sensitive(true);

            headers.insert(header::AUTHORIZATION, auth);

            headers
        })
        .body(body)
        .send()
        .await?;

    if res.status().as_u16() >= 400 {
        let body = res.text().await.unwrap_or_default();
        eyre::bail!("Ntfy server returned error: {}", body);
    }

    Ok(())
}

#[tracing::instrument(skip_all)]
pub async fn run(event_tx: Sender<Event>, mut ntfy_rx: UnboundedReceiver<Message>) -> Result<()> {
    let ntfy_client = Client::builder().user_agent(APP_USER_AGENT).build()?;

    while let Some(message) = ntfy_rx.recv().await {
        match message {
            Message::Ntfy(data) => {
                tracing::trace!(topic = data.topic, "Sending Ntfy notification");
                if let Err(err) = send_ntfy(&ntfy_client, &data).await {
                    tracing::error!("Failed to send to Ntfy: {:?}", err);
                    event_tx.send(Event::error(data.topic, err))?;
                }
            }
        }
    }

    tracing::debug!("Stopped receiving messages");
    Ok(())
}

#[tracing::instrument(skip_all)]
pub fn worker(event_tx: Sender<Event>, ntfy_rx: UnboundedReceiver<Message>) -> Result<()> {
    let rt = runtime::Builder::new_current_thread()
        .thread_name("sender")
        .enable_io()
        .build()?;

    rt.block_on(run(event_tx, ntfy_rx))
}
